private import std.stdio;
private import std.string;
private import std.math;
private import util_win.mainwindow;
private import util_win.clipboard;
private import util_win.dib;
private import util.expression_real;


char[] real2str(real r, int imax, real log10min=-5, real log10max=19)
{
    int  i, j;
    real x, y, u;
    char[] exp;
    
    char[] str = new char[imax];

    if (r==real.nan)  { str[0..3]="nan";  return str[0..3]; }
    if (r>=real.max)  { str[0..3]="inf";  return str[0..3]; }
    if (r<=-real.max) { str[0..4]="-inf"; return str[0..4]; }
    if (-real.min<=r && r<=real.min) { str[0..1]="0"; return str[0..1]; }

    i = 0;    
    if (r<0) { r=-r; str[i++]='-'; }
    
    real log10r = floor(log10(r));  // considering; floor(log10(1e-100))=-101
    int  ilog10r = cast(int)log10r;

    if (log10r < log10min || log10max < log10r) {
        r *= pow(10, -log10r);
        exp=format("e%d", ilog10r);
        imax-=exp.length;
        ilog10r=0;
    } else {
        if (ilog10r<0) ilog10r=0;
    }

    for(u=1,j=i+ilog10r+2;j<imax;j++){ u*=0.1; }
    r += u*0.5;
    
    for(x=r,j=ilog10r;j>=0;j--) { y=floor(x*0.1); str[j+i]='0'+cast(int)(x-y*10); x=y; }
    r -= floor(r);
    i += ilog10r+1;
    if (i<imax && r>u) {
        str[i++]='.';
        for(x=r*10;i<imax;x*=10) { y=floor(x); str[i++]='0'+cast(int)y; x-=y; }
        while(str[--i]=='0') { str[i]=0; }
        i++;
    }

    str[i..i+exp.length] = exp[];
    
    return str;
}



//--------------------------------------------------------------------------------
// Screen
//--------------------------------------------------------------------------------
class Screen : CDIBSurface
{
public:
    CDIBSurface  m_backsurf;
    CDIBSurface  m_graphsurf;
    char[][]     m_text;
    int          crsl_x, crsl_y, topline;
    int          text_cx, text_cy;
    
    this(int w, int h)
    {
        text_cx = w;
        text_cy = h;
        m_backsurf  = new CDIBSurface("font.bmp");
        m_graphsurf = new CDIBSurface(w * 6, h * 8);
        super(m_graphsurf.cx, m_graphsurf.cy);
        
        cls();
    }
    
    void cls()
    {
        m_graphsurf[] = 0x009f7f;
        m_text.length = text_cy;
        m_text[] = null;
        crsl_x = crsl_y = 0;
        topline = 0;
    }

    void print(int x, int y, char[] str)
    {
        if (x<0 || x>=text_cx || y<0 || y>=text_cy) return;
        crsl_x = x;
        crsl_y = y;
        print(str);
    }
    
    void print(char[] str)
    {
        char[][] ln = splitlines(str);
        char[]   sp = repeat(" ", crsl_x);
        foreach(s; ln) {
            if (crsl_y == text_cy) scroll();
            m_text[crsl_y+topline] = sp ~ s;
            crsl_y++;
        }
    }
    
    void scroll()
    {
        topline++;
        crsl_y--;
        while (topline+text_cy>=m_text.length) { m_text.length = m_text.length << 1; }
    }

    
    void gprint(int x, int y, ubyte[] pat, uint col=0)
    {
        ubyte r = col;
        ubyte g = col >> 8;
        ubyte b = col >> 16;
        foreach (int i, ubyte p; pat) {
            if (cy<y+i) break;
            uint add = address(x+8, y+i);
            while (p != 0) {
                if (p & 1) {
                    m_graphsurf.bit[add+2] = r;
                    m_graphsurf.bit[add+1] = g;
                    m_graphsurf.bit[add]   = b;
                }
                add-=3;
                p>>=1;
            }
        }
    }
    
    static ubyte[] convert(char[] hex)
    {
        ubyte[] pat = new ubyte[hex.length >> 1];
        int chi;
        for(int i=0; i<pat.length; i++) {
            chi = i<<1;
            pat[i] = (ctob(hex[chi])<<4) | ctob(hex[chi+1]);
        }
        return pat;
    }
    
    static ubyte ctob(char c) { return (c<='9') ? (c-'0') : (c-'7'); }
    
    
    char[] get_line(int idx)
    {
        return m_text[topline+idx];
    }
    
    void line(int idx)
    {
        static uint line(uint col) { return 0x000000; }
        calc(0, idx*8+7, cx, 1, &line);
    }
    
    void _draw_text()
    {
        bit[] = m_graphsurf.bit[];
        
        int y = 0;
        foreach(s; m_text[topline..topline+text_cy]) {
            int x = 0;
            foreach(c; s) {
                copy(x, y, 6, 8, m_backsurf, (c&0x0f)<<3, (c&0xf0)>>1);
                if ((x+=6) > cx) break;
            }
            y+=8;
        }
    }
}




//--------------------------------------------------------------------------------
// Main
//--------------------------------------------------------------------------------
class MainApp : TMainWindow!(MainApp)
{
// Member value
private:
    CClipboardViewer  m_viewer;
    CExpressionReal   m_exp;
    Screen            m_screen;
    bool              m_lock;
    int               m_mousex, m_mousey;

// Constructor
public:
    this()
    {
        m_viewer = new CClipboardViewer();
        m_exp = new CExpressionReal();
        m_screen = new Screen(20,10);
        m_screen[] = 0x009f7f;
    }
    

// Call back functions
public:
    void enterInstance(char[][] arg)
    {
        m_setting.style = WS_VISIBLE | WS_POPUP | WS_BORDER;
        setWindow("clicom", m_screen.cx+8, m_screen.cy+8);
    }
    
    
    void exitInstance()
    {
    }
    
    
    bool proc(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
    {
        HDC         wndDC, bitmapDC;
        PAINTSTRUCT ps;
        int         x, y, f;
        
        static bool  hold = false;
        static POINT drag_pt;

        switch (WM) {
        case WM_CREATE:
            m_lock = false;
            m_viewer.begin(getHandle(), &onDrawClipboard);
            setTopmost();
            return true;
        case WM_DESTROY:
            m_viewer.end();
            return true;
        case WM_PAINT:
            wndDC = BeginPaint(getHandle(), &ps);
            bitmapDC = CreateCompatibleDC(wndDC);
            SelectObject(bitmapDC, m_screen.getHBITMAP());
            onPaint(bitmapDC);
            BitBlt(wndDC, 4, 4, m_screen.cx, m_screen.cy, bitmapDC, 0, 0, SRCCOPY);
            DeleteDC(bitmapDC);
            EndPaint(getHandle(), &ps);
            return true;
        case WM_MOUSEMOVE:
            setCapture();
            parseMouseEvent(wParam, lParam, x, y, f);
            m_mousex = x;
            m_mousey = y;
            if (hold) { toScreen(x, y); move(x-drag_pt.x, y-drag_pt.y); }
            else if (!isInClient(x, y)) { m_mousey=0; releaseCapture(); }
            invalidateRect(false);
            return true;
        case WM_LBUTTONDOWN:
            parseMouseEvent(wParam, lParam, x, y, f);
            drag_pt.x = x;
            drag_pt.y = y;
            hold = true;
            return true;
        case WM_LBUTTONUP:
            parseMouseEvent(wParam, lParam, x, y, f);
            y = (y - 4) >> 3;
            if (y>=0 && y<m_screen.text_cy) {
                char[] str = m_screen.get_line(y);
                if (str.length > 0) {
                    m_lock = true;
                    setClipboard(str, CF_TEXT);
                    m_lock = false;
                }
            }
            hold = false;
            return true;
        case WM_RBUTTONUP:
            parseMouseEvent(wParam, lParam, x, y, f);
            PostQuitMessage(0);
            return true;
        default:
            return m_viewer.proc(hWnd, WM, wParam, lParam);
        }
    }


// Callback from clipboard viewer
public:
    void onDrawClipboard()
    {
        if (m_lock) return;

        char[]   clip = getClipboard(CF_TEXT);
        if (clip.length == 0) return;
        
        char[][] line = splitlines(clip);
        if (!m_exp.parse(line[0])) {
debug writefln("failed to parse");
            return;
        }
        
        char[] str = real2str(m_exp.exec(), 20);
        m_screen.print(str);
        invalidateRect();
    }
    
    
    bool onPaint(HDC hDC)
    {
        m_screen._draw_text();
        int y = (m_mousey-4) >> 3;
        if (y>=0 && y<m_screen.text_cy) {
            m_screen.line(y);
        }
        return true;
    }
}

