import std.stdio;
import std.file;


// Bitmap Info Header 
//--------------------------------------------------
struct BIH {
    uint    biSize = 40;
    int     biWidth;
    int     biHeight;
    ushort  biPlanes;
    ushort  biBitCount;
    uint    biCompression;
    uint    biSizeImage;
    int     biXPelsPerMeter;
    int     biYPelsPerMeter;
    uint    biClrUsed;
    uint    biClrImportant;
} 


// read limited information of bitmap
ubyte[] readBitmap(char[] nameFile, out int width, out int height, out int bit, out ubyte[] pixel, out ubyte[] tableColor)
{
    ubyte[] data = cast(ubyte[])read(nameFile);
    if (data.length < 54) return null;
   
    // read header
    ubyte[] header = data[0..14];
    ubyte[] info   = data[14..54];
    if (header[0]!='B' || header[1]!='M')   return null;
    if ((cast(BIH*)info).biCompression != 0) return null;

    // get information
    width  = (cast(BIH*)info).biWidth;
    height = (cast(BIH*)info).biHeight;
    bit    = (cast(BIH*)info).biBitCount;

    // refer color table
    uint cntColor = 0;
    if (bit==8 || bit==1) {
        cntColor = (cast(BIH*)info).biClrUsed;
        if (cntColor==0) cntColor=1<<bit;
        if (data.length < 54+cntColor) return null;
        tableColor = data[54..54+(cntColor<<2)];
    }

    pixel = data[54+(cntColor<<2)..length];
    return data;
}


// convert
bool convertBitmap24(out ubyte[] pix24, ubyte[] pixel, ubyte[] tableColor, int bit)
{
    uint  add, i;
    ubyte f;
    
    switch (bit) {
    case 24:
        pix24.length = pixel.length;
        pix24[] = pixel;
        return true;
    case 8:
        pix24.length = pixel.length * 3;
        add = 0;
        foreach(p; pixel) {
            p <<= 2;
            pix24[add++] = tableColor[p];
            pix24[add++] = tableColor[p+1];
            pix24[add++] = tableColor[p+2];
        }
        return true;
    case 1:
        pix24.length = pixel.length * 24;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++) {
                f = (p&128)>>5;
                pix24[add++] = tableColor[f];
                pix24[add++] = tableColor[f+1];
                pix24[add++] = tableColor[f+2];
                p <<= 1;
            }
        }
        return true;
    default:
        return false;
    }
}


// convert
void convertBitmap32(out ubyte[] pix32, ubyte[] pixel, ubyte[] tableColor, int bit)
{
    uint  add, i;
    ubyte f;
    
    switch (bit) {
    case 24:
        pix32.length = (pixel.length << 2)/ 3;
        add = 0;
        for (i=0; i<pix32.length; i+=4) {
            pix32[i]   = pixel[add++];
            pix32[i+1] = pixel[add++];
            pix32[i+2] = pixel[add++];
            pix32[i+3] = 255;
        }
        break;
    case 8:
        pix32.length = pixel.length * 4;
        add = 0;
        foreach(p; pixel) {
            p <<= 2;
            pix32[add++] = tableColor[p];
            pix32[add++] = tableColor[p+1];
            pix32[add++] = tableColor[p+2];
            pix32[add++] = tableColor[p+3];
        }
        break;
    case 1:
        pix32.length = pixel.length * 32;
        add = 0;
        foreach(p; pixel) {
            for (i=0; i<8; i++) {
                f = (p&128)>>5;
                pix32[add++] = tableColor[f];
                pix32[add++] = tableColor[f+1];
                pix32[add++] = tableColor[f+2];
                pix32[add++] = tableColor[f+3];
                p <<= 1;
            }
        }
        break;
    default:
        break;
    }
}
