module util_win.clipboard;

private import std.outofmemory;
private import std.c.stdio;
private import win32.windows;
private import util_win.window;

/*
// クリップボードフォーマット
enum {
    CF_TEXT = 1,
    CF_BITMAP,
    CF_METAFILEPICT,
    CF_SYLK,
    CF_DIF,
    CF_TIFF,
    CF_OEMTEXT,
    CF_DIB,
    CF_PALETTE,
    CF_PENDATA,
    CF_RIFF,
    CF_WAVE,
    CF_UNICODETEXT,
    CF_ENHMETAFILE,
    CF_HDROP,
    CF_LOCALE,
    CF_MAX, // = 17
    CF_OWNERDISPLAY   = 128,
    CF_DSPTEXT,
    CF_DSPBITMAP,
    CF_DSPMETAFILEPICT, // = 131
    CF_DSPENHMETAFILE = 142,
    CF_PRIVATEFIRST   = 512,
    CF_PRIVATELAST    = 767,
    CF_GDIOBJFIRST    = 768,
    CF_GDIOBJLAST     = 1023
}
*/


// クリップボードにデータを設定
bool setClipboard(char[] data, uint format)
{
    // クリップボードを開ける
    if (!OpenClipboard(null)) return false;
    
    // クリップボードの内容の削除
    EmptyClipboard();
    
    // ヒープ領域の確保
    int size = data.length;
    
    HGLOBAL hGlobal = GlobalAlloc(GHND, size + 1);
    if (hGlobal == null) throw new OutOfMemoryException;
    
    // ヒープ領域のロック
    char* pHeap = cast(char*)GlobalLock(hGlobal);
    if (!pHeap) {
        GlobalFree(hGlobal);
        throw new OutOfMemoryException;
    }

    // 書き込み
    char[] heap = pHeap[0..size];
    heap[] = data[];
    
    // ロックの解除
    GlobalUnlock(hGlobal);
    
    // クリップボードに登録
    SetClipboardData(format, hGlobal);

    // クリップボードを閉じる
    scope(exit) CloseClipboard();
    
    return true;
}


// クリップボードからデータを取得
char[] getClipboard(uint format)
{
    char[] data;
    
    // フォーマットチェック
    if (!IsClipboardFormatAvailable(format)) return data;
    
    // クリップボードを開ける
    if (!OpenClipboard(null)) return data;

    // クリップボード内容のグローバルハンドルの取得
    HGLOBAL hGlobal = GetClipboardData(format);                        
    if (hGlobal == null) throw new OutOfMemoryException;
        
    // ヒープ領域のロック
    char* pHeap = cast(char*)GlobalLock(hGlobal);
    if (pHeap == null) return data;
        
    // データサイズの取得/読込み
    int size = GlobalSize(hGlobal) - 1;
    data = new char[size];
    data[] = pHeap[0..size];

    //ロックの解除
    GlobalUnlock(hGlobal);

    // クリップボードを閉じる
    scope(exit) CloseClipboard();

    return data;
}



// クリップボードビュアチェーンクラス
class CClipboardViewer : IWindow
{
private:
    HWND            m_hNextCBVChain = null;
    void delegate() m_dgCBVCallBack = null;

public:
    this()
    {
    }
    
    
public:
    // クリップボード監視開始．クリップボード内容が変更されたら dg() をコールバック．
    void begin(HWND hWnd, void delegate() dg)
    {
        formHandle(hWnd);
        m_hNextCBVChain = SetClipboardViewer(hWnd);
        m_dgCBVCallBack = dg;
    }


    // クリップボードの監視中は window procedure からこの関数を呼び出す．
    bool proc(HWND hWnd, uint WM, WPARAM wParam, LPARAM lParam)
    {
        if (hWnd != m_hWnd) return false;
        
        switch (WM) {
        case WM_CHANGECBCHAIN:
            if (m_hNextCBVChain == cast(HWND)wParam) m_hNextCBVChain = cast(HWND)lParam;
            if (m_hNextCBVChain) SendMessage(m_hNextCBVChain, WM_CHANGECBCHAIN, wParam, lParam);
            return true;
        case WM_DRAWCLIPBOARD:
            if (m_dgCBVCallBack) m_dgCBVCallBack();
            SendMessage(m_hNextCBVChain, WM_DRAWCLIPBOARD, 0, 0);
            return true;
        default:
            return false;
        }
    }


    // クリップボード監視終了．終了時必ずこれを呼ぶ．
    void end()
    {
        ChangeClipboardChain(m_hWnd, m_hNextCBVChain);
    }
}
