
import std.stdio;
import win32.windows;
import util.bitmap;
import util_win.setstruct;

pragma(lib, "gdi32");


// Wrapper for DIB Section
class CDIBSurface
{
protected:
    BITMAPINFO  bmi;            // Bitmap Info
    HBITMAP     hdib = null;    // Bitmap Hndle

public:
    ubyte[]     bit = null;     // pixel data
    uint        cx = 0;         // width
    uint        cy = 0;         // height
    uint        sizeLine = 0;   // [byte/line]

public:
    this()
    {
        setBMI(&(bmi.bmiHeader), null, 0, 0, 24);
    }

    this(int w, int h)
    {
        this();
        alloc(w, h);
    }

    this(HDC hCompDC, int w, int h)
    {
        this();
        alloc(hCompDC, w, h);
    }

    this(char[] nameFile)
    {
        this();
        load(nameFile);
    }

    ~this()
    {
        free();
    }

    // allocate memory
    bool alloc(int w, int h)
    {
        HDC hDesktop = CreateDC("DISPLAY", null, null, null);
        return alloc(hDesktop, w, h);
    }
    
    bool alloc(HDC hCompDC, int w, int h)
    {
        free();
        
        sizeLine = setBMI(&(bmi.bmiHeader), null, w, h, 24);
        ubyte* buf;
        hdib = CreateDIBSection(hCompDC, &bmi, DIB_RGB_COLORS, cast(void**)(&buf), null, 0);
        if (hdib == null) return false;
        cx = w;
        cy = h;
        bit = buf[0..sizeLine*h];
        
        return true;
    }

    // free all
    void free()
    {
        if (hdib != null) DeleteObject(hdib);
        setBMI(&(bmi.bmiHeader), null, 0, 0, 24);
        hdib = null;
        bit  = null;
    }
    
    // load
    bool load(char[] nameFile)
    {
        int bmpbit, w, h;
        auto ubyte[] pxl;
        auto ubyte[] tbl;
        if (readBitmap(nameFile, w, h, bmpbit, pxl, tbl) == null) return false;
        if (!alloc(w, h)) return false;
        if (!convertBitmap24(bit, pxl, tbl, bmpbit)) return false;
        return true;
    }

    void copy(int x, int y, int w, int h, CDIBSurface src, int sx, int sy)
    {
        uint dadd = address(x,y);
        uint sadd = src.address(sx,sy);
        uint wadd = w*3;
        for (int i=0; i<h; i++) {
            bit[dadd..dadd+wadd] = src.bit[sadd..sadd+wadd];
            dadd -= sizeLine;
            sadd -= src.sizeLine;
        }
    }
    
    void calc(int x, int y, int w, int h, uint function(uint) fn)
    {
        uint addmin, addmax, add, col;

        for (int i=0; i<h; i++) {
            addmin = address(x, y+i);
            addmax = addmin + w*3;
            add = addmin;
            while (add<addmax) {
                col = (bit[add]<<16) | (bit[add+1]<<8) | bit[add+2];
                col = fn(col);
                bit[add++] = (col >> 16) & 0xff;
                bit[add++] = (col >> 8)  & 0xff; 
                bit[add++] = col & 0xff; 
            }
        }
    }

    // calculate address
    uint address(uint x, uint y)
    {
        return (cy - y - 1) * sizeLine + x * 3;
    }

    // address
    uint opIndex(uint x, uint y)
    {
        uint add = address(x, y);
        return (bit[add]) | (bit[add+1]<<8) | (bit[add+2]<<16);
    }

    // address
    ubyte[] opIndex(uint y)
    {
        uint add = address(0, y);
        return bit[add..sizeLine];
    }

    // address
    uint opIndexAssign(uint col, uint x, uint y)
    {
        uint add = address(x, y);
        bit[add+2] = col;
        bit[add+1] = col >> 8;
        bit[add]   = col >> 16;
        return col;
    }

    // address
    ubyte[] opIndexAssign(uint col, uint y)
    {
        ubyte r = col;
        ubyte g = col >> 8;
        ubyte b = col >> 16;
        uint addmin = address(0, y);
        uint addmax = addmin + sizeLine;
        uint add=addmin;
        while (add<addmax) {
            bit[add++] = b;
            bit[add++] = g;
            bit[add++] = r;
        }
        return bit[addmin..addmin+sizeLine];
    }
    
    // slice
    ubyte[] opSlice()
    {
        return bit[];
    }
    
    // slice
    ubyte[] opSliceAssign(uint col)
    {
        for (uint y=0; y<cy; y++) { this[y] = col; }
        return bit[];
    }
    
    // HBITMAP
    HBITMAP getHBITMAP()
    {
        return hdib;
    }
};



